#include "ehs/io/socket/ehc/NetFrags.h"

namespace ehs
{
	NetFrags::~NetFrags()
	{
		delete[] data;
	}

	NetFrags::NetFrags()
		: data(nullptr), size(0)
	{
	}

	NetFrags::NetFrags(const Header &header, const Serializer<UInt_64> &payload)
		: header(header), data(new Serializer<UInt_64>[header.fragments]), size(header.fragments)
	{
        this->header.fragment = 0;
		data[header.fragment] = payload;
	}

	NetFrags::NetFrags(const Header &header, const UInt_64 size)
		: header(header), data(new Serializer<UInt_64>[size]), size(size)
	{
        this->header.fragments = size;
        this->header.fragment = 0;
	}

	NetFrags::NetFrags(NetFrags &&frags) noexcept
		: header(frags.header), data(frags.data), size(frags.size)
	{
		frags.header = {};
		frags.data = nullptr;
		frags.size = 0;
	}

	NetFrags::NetFrags(const NetFrags &frags)
		: header(frags.header), data(new Serializer<UInt_64>[frags.size]), size(frags.size)
	{
		for (UInt_64 i = 0; i < size; ++i)
			data[i] = frags.data[i];
	}

	NetFrags &NetFrags::operator=(NetFrags &&frags) noexcept
	{
		if (this == &frags)
			return *this;

		header = frags.header;

		delete[] data;
		data = frags.data;

		size = frags.size;

		frags.header = {};
		frags.data = nullptr;
		frags.size = 0;

		return *this;
	}

	NetFrags &NetFrags::operator=(const NetFrags &frags)
	{
		if (this == &frags)
			return *this;

        header = frags.header;
		delete[] data;
		data = new Serializer<UInt_64>[frags.size];
		for (UInt_64 i = 0; i < frags.size; ++i)
			data[i] = frags.data[i];
		size = frags.size;

		return *this;
	}

	NetFrags::operator Serializer<UInt_64> *() const
	{
		return data;
	}

	Header NetFrags::GetHeader() const
	{
		return header;
	}

	UInt_64 NetFrags::Size() const
	{
		return size;
	}

	bool NetFrags::IsComplete() const
	{
		for (UInt_64 i = 0; i < size; ++i)
			if (!data[i].Size())
				return false;

		return true;
	}

	Packet NetFrags::Combine() const
	{
        UInt_64 rSize = 0;
        for (UInt_64 i = 0; i < size; ++i)
            rSize += data[i].Size();

		Packet result =
		{
		    header,
            {Endianness::LE, rSize}
		};
		result.header.fragments = 0;

        for (UInt_64 i = 0; i < size; ++i)
            result.payload.WriteSer(data[i]);

        result.payload.SetOffset(0);

		return result;
	}
}